from base.fields import fields

class cross_reference(fields):

    def validateTargetApplication(self):

        if self.swimlane_target.validateApp(byID = self.field_obj['targetId']) == False:
            self.logger.info('Cross Reference Target ID not found in Target: {}'.format(self.field_obj['targetId']))
            source_app_name = self.swimlane_source.validateApp(byID = self.field_obj['targetId'])
            if source_app_name == False:
                self.logger.error('FAILED TO FIND: Orphan Cross Reference APP: {}'.format(source_app_name))
                return False
            else:
                self.logger.info('Orphan Cross Reference APP: {}'.format(source_app_name))
                target_app_id = self.swimlane_target.validateApp(byName = source_app_name)

                if target_app_id != False:
                    self.logger.debug('Cross Reference App ID found in target: {}'.format(target_app_id))
                    self.field_obj['targetId'] = target_app_id
                    self.updates_made = True
                else:
                    self.logger.error('FAILED TO FIND: Target Application ID')
                    return False
        else:
            self.logger.debug('Cross Reference Field {} already valid'.format(self.field_obj['name']))

        return True

    def validateFieldColumns(self):
        app_name = self.swimlane_target.validateApp(byID = self.field_obj['targetId'])

        if app_name != False:
            source_ref_app_fields  = self.swimlane_source.getFieldDictinoary(byName=app_name)
            target_ref_app_fields = self.swimlane_target.getFieldDictinoary(byName=app_name)
            new_columns = []
            for ref_field in self.field_obj['columns']:
                if self.swimlane_target.validateField(target_ref_app_fields, byID=ref_field) == False:
                    ref_source_app_name = self.swimlane_source.validateField(source_ref_app_fields, byID=ref_field)
                    target_field_id = self.swimlane_target.validateField(target_ref_app_fields, byName=ref_source_app_name)
                    if target_field_id != False:
                        self.logger.info('Field {} migrated'.format(ref_source_app_name))
                        new_columns.append(target_field_id)
                        self.updates_made = True
                    else:
                        self.logger.error('CROSS REFERENCE COLUMNS: Field ID {} failed to find a match'.format(ref_field))
                        self.failed_fields.append(ref_field)
                else:
                    self.logger.debug('Field {} already valid'.format(self.swimlane_target.validateField(target_ref_app_fields, byID=ref_field)))
                    new_columns.append(ref_field)

            if len(new_columns) == len(self.field_obj['columns']):
                self.logger.debug('All Columns Migrated count: {}'.format(len(new_columns)))
                self.field_obj['columns'] = new_columns
                return True
            else:
                self.logger.error('FAILED TO MIGRATE COLUMNS: {}'.format(self.failed_fields))
                return False
        else:
            self.logger.critical('Cross Reference Field "{}" target application ID: "{}" could not be found'.format(self.field_obj['name'], self.field_obj['targetId']))
            return False



import json, sys
from copy import copy
from swimlane_migrator.usergroups.groups import groups
from swimlane_migrator.utils import find_in_list

class report(object):

    def __init__(self, config):
        self.mode = config['mode']
        self.groups = groups(copy(config))

        self.logger = config.pop('logger', None)
        self.swimlane_source = config.pop('swimlane_source', None)
        self.swimlane_target = config.pop('swimlane_target', None)

        if self.swimlane_source is None:
            raise  Exception('You must provide a swimlane_source object')

        if self.swimlane_target is None:
            raise  Exception('You must provide a swimlane_target object')

        self.remove_keys = ['createdDate', 'modifiedDate', 'createdByUser', 'modifiedByUser']

        self.migratedReports = []


    def addReport(self, swimlane_server, report):
        report_response = swimlane_server.swimlane.request(
            'post',
            '/reports',
            json = report
        )

        return json.loads(report_response.text)

    def updateReport(self, swimlane_server, report_id, report):
        report_response = swimlane_server.swimlane.request(
            'put',
            '/reports/{}'.format(report_id),
            json = report
        )

        return json.loads(report_response.text)

    def getReportsByApp(self, swimlane_server, appID):
        report_response = swimlane_server.swimlane.request(
            'get',
            '/reports/app/{}'.format(appID)
        )

        return json.loads(report_response.text)

    def getReportsByWorkspace(self, swimlane_server, workspaceID):
        report_response = swimlane_server.swimlane.request(
            'get',
            '/reports/workspace/{}'.format(workspaceID)
        )

        return json.loads(report_response.text)

    def getReportsByApp(self, swimlane_server, appID):
        report_response = swimlane_server.swimlane.request(
            'get',
            '/reports/app/{}'.format(appID)
        )

        return json.loads(report_response.text)

    def getReportByID(self, swimlane_server, reportID):
        report_response = swimlane_server.swimlane.request(
            'get',
            '/reports/{}'.format(reportID)
        )

        return json.loads(report_response.text)

    def getReports(self, swimlane_server):
        report_response = swimlane_server.swimlane.request(
            'get',
            '/reports'
        )

        return json.loads(report_response.text)

    def getReportDictionary(self, swimlane_server, appID):
        report_response = self.getReportsByApp(swimlane_server, appID)

        reports_dictionary = {}
        for report in report_response:
            reports_dictionary[report['name']] = report

        return reports_dictionary

    def validateApplications(self, application_list):
        new_application_list = []
        for application_id in application_list:
            source_app_name = self.swimlane_source.validateApp(byID=application_id)
            if source_app_name != False:
                target_app_id = self.swimlane_target.validateApp(byName=source_app_name)
                if target_app_id != False:
                    new_application_list.append(target_app_id)

        return new_application_list

    def fixColumns(self, columns, source_app_id, target_app_id):
        new_columns = []

        source_field_dictionary= self.swimlane_source.getFieldDictinoary(byID=source_app_id)
        target_field_dictionary= self.swimlane_target.getFieldDictinoary(byID=target_app_id)

        for column in columns:

            source_field_name = self.swimlane_source.validateField(source_field_dictionary, byID=column)
            if source_field_name == False:
                self.logger.error('Report Columns -- Source Field ID {} not found'.format(column))
                continue

            target_field_id = self.swimlane_target.validateField(target_field_dictionary, byName=source_field_name)
            if target_field_id == False:
                self.logger.error('Report Columns -- Target Field Name {} not found in {}'.format(source_field_name, target_field_dictionary))
                continue

            new_columns.append(target_field_id)

        return new_columns

    def fixSorts(self, sorts, source_app_id, target_app_id):
        new_sorts = {}
        new_sorts['$type'] = sorts['$type']

        source_field_dictionary= self.swimlane_source.getFieldDictinoary(byID=source_app_id)
        target_field_dictionary= self.swimlane_target.getFieldDictinoary(byID=target_app_id)

        for field_id, sort in sorts.items():
            if field_id != '$type':

                source_field_name = self.swimlane_source.validateField(source_field_dictionary, byID=field_id)
                if source_field_name == False:
                    self.logger.error('Report Sort -- Source Field ID {} not found'.format(field_id))
                    continue

                target_field_id = self.swimlane_target.validateField(target_field_dictionary, byName=source_field_name)
                if target_field_id == False:
                    self.logger.error('Report Sort -- Target Field Name {} not found in {}'.format(source_field_name, target_field_dictionary))
                    continue

                new_sorts[target_field_id] = sort

        return new_sorts

    def fixFilters(self, filters, source_app_id, target_app_id):
        new_filter = []

        source_field_dictionary= self.swimlane_source.getFieldDictinoary(byID=source_app_id)
        target_field_dictionary= self.swimlane_target.getFieldDictinoary(byID=target_app_id)

        for report_filter in filters:
            if 'fieldId' in report_filter:
                source_field_name = self.swimlane_source.validateField(source_field_dictionary, byID=report_filter['fieldId'])
                if source_field_name == False:
                    self.logger.error('Report Filter -- Source Field ID {} not found'.format(report_filter['fieldId']))
                    continue

                target_field_id = self.swimlane_target.validateField(target_field_dictionary, byName=source_field_name)
                if target_field_id == False:
                    self.logger.error('Report Filter -- Target Field Name {} not found in {}'.format(source_field_name, target_field_dictionary))
                    continue

                report_filter['fieldId'] = target_field_id
                new_filter.append(report_filter)

            if 'value' in report_filter:
                if isinstance(report_filter['value'], dict):
                    if '$type' in report_filter['value']:
                        if report_filter['value']['$type'] == 'Core.Models.Utilities.UserGroupSelection, Core':
                            group_obj = self.groups.lookupGroup(self.swimlane_target, report_filter['value']['name'])
                            if len(group_obj) == 1:
                                report_filter['value']['id'] = group_obj[0]['id']


        return new_filter

    def fixgroupBys(self, groupBys, source_app_id, target_app_id):
        new_groupBy = []

        source_field_dictionary= self.swimlane_source.getFieldDictinoary(byID=source_app_id)
        target_field_dictionary= self.swimlane_target.getFieldDictinoary(byID=target_app_id)

        for groupBy in groupBys:
            if 'fieldId' in groupBy:
                source_field_name = self.swimlane_source.validateField(source_field_dictionary, byID=groupBy['fieldId'])
                if source_field_name == False:
                    self.logger.error('Report GroupBy -- Source Field ID {} not found'.format(groupBy['fieldId']))
                    continue

                target_field_id = self.swimlane_target.validateField(target_field_dictionary, byName=source_field_name)
                if target_field_id == False:
                    self.logger.error('Report GroupBy -- Target Field Name {} not found in {}'.format(source_field_name, target_field_dictionary))
                    continue

                groupBy['fieldId'] = target_field_id
                new_groupBy.append(groupBy)

        return new_groupBy

    def fixAggregates(self, aggregates, source_app_id, target_app_id):
        new_aggregates = []

        source_field_dictionary= self.swimlane_source.getFieldDictinoary(byID=source_app_id)
        target_field_dictionary= self.swimlane_target.getFieldDictinoary(byID=target_app_id)

        for aggregate in aggregates:
            if 'fieldId' in aggregate:
                source_field_name = self.swimlane_source.validateField(source_field_dictionary, byID=aggregate['fieldId'])
                if source_field_name == False:
                    self.logger.error('Report Aggregates -- Source Field ID {} not found'.format(aggregate['fieldId']))
                    continue

                target_field_id = self.swimlane_target.validateField(target_field_dictionary, byName=source_field_name)
                if target_field_id == False:
                    self.logger.error('Report Aggregates -- Target Field Name {} not found in {}'.format(source_field_name, target_field_dictionary))
                    continue

                aggregate['fieldId'] = target_field_id
                new_aggregates.append(aggregate)

        return new_aggregates

    def fixPermissions(self, permissions):
        new_permissions = {}

        for key, permission in permissions.items():
            if key != '$type':
                if permission['type'] == 'Role':
                    target_role_id = self.swimlane_target.validateRole(byName=permission['name'])
                    permission['id'] = target_role_id
                    new_permissions[target_role_id] = permission
            else:
                new_permissions[key] = permission

        return new_permissions

    def repairReport(self, report_obj):
        source_app_id = None
        if len(report_obj['applicationIds']) > 0:
            source_app_id = report_obj['applicationIds'][0]

            report_obj['applicationIds'] = self.validateApplications(report_obj['applicationIds'])
            if len(report_obj['applicationIds']) > 0:
                target_app_id = report_obj['applicationIds'][0]

                report_obj['columns'] = self.fixColumns(report_obj['columns'], source_app_id, target_app_id)
                report_obj['sorts'] = self.fixSorts(report_obj['sorts'], source_app_id, target_app_id)
                report_obj['filters'] = self.fixFilters(report_obj['filters'], source_app_id, target_app_id)
                report_obj['groupBys'] = self.fixgroupBys(report_obj['groupBys'], source_app_id, target_app_id)
                report_obj['aggregates'] = self.fixAggregates(report_obj['aggregates'], source_app_id, target_app_id)
                report_obj['permissions'] = self.fixPermissions(report_obj['permissions'])
            else:
                self.logger.critical('Report {} was unable to migrate, no applications assigned to report'.format(report_obj['name']))
                return None

        return report_obj


    def migrateReports(self, mode,byWorkspace = None, byApp = None,):

        source_reports = None
        target_reports = None

        if byWorkspace is not None:
            source_reports = self.getReportsByWorkspace(self.swimlane_source, byWorkspace['source'])
            target_reports = self.getReportsByWorkspace(self.swimlane_target, byWorkspace['target'])

        if byApp is not None:
            source_reports = self.getReportsByApp(self.swimlane_source, byApp['source'])
            target_reports = self.getReportsByApp(self.swimlane_target, byApp['target'])


        if source_reports is None and target_reports is None:
            raise Exception('Unable to load reports by Workstation ID: {} '.format(byWorkspace))

        self.logger.info('Count of Source Reports: {}'.format(len(source_reports)))

        for report in source_reports:
            if report['id'] not in self.migratedReports:
                self.logger.info('### Working on Report {}'.format(report['name']))
                self.migratedReports.append(report['id'])
                target_report = copy(report)
                target_report = self.repairReport(target_report)
                if target_report is not None:
                    target_report_key = find_in_list(target_reports, 'name', report['name'], self.logger)
                    if target_report_key is None:
                        for key in self.remove_keys:
                            if key in target_report:
                                target_report.pop(key)
                        if mode == 'Live':
                            try:
                                new_report = self.addReport(self.swimlane_target, target_report)
                            except Exception as e:
                                self.logger.critical('Failed to add Report "{}" Error: {}'.format(target_report['name'], e))
                    else:
                        update_report = target_reports[target_report_key]
                        update_report['applicationIds'] = copy(target_report['applicationIds'])
                        update_report['columns'] = copy(target_report['columns'])
                        update_report['sorts'] = copy(target_report['sorts'])
                        update_report['filters'] = copy(target_report['filters'])
                        update_report['groupBys'] = copy(target_report['groupBys'])
                        update_report['aggregates'] = copy(target_report['aggregates'])
                        update_report['permissions'] = copy(target_report['permissions'])
                        if mode == "Live":
                            try:
                                self.updateReport(self.swimlane_target, update_report['id'], update_report)
                            except Exception as e:
                                self.logger.critical('Failed to Update Report "{}" Error: {}'.format(update_report['name'], e))



import json
from StringIO import StringIO
from copy import copy
from swimlane_migrator.utils import buildLogDictionary

class tasksEngine():

    def __init__(self, config):

        self.logger = config.pop('logger', None)
        self.swimlane_source = config.pop('swimlane_source', None)
        self.swimlane_target = config.pop('swimlane_target', None)

        if self.logger is None:
            raise  Exception('You must provide a logger object')

        if self.swimlane_source is None:
            raise  Exception('You must provide a swimlane_source object')

        if self.swimlane_target is None:
            raise  Exception('You must provide a swimlane_target object')

        self.log_dictionary = self.swimlane_target.log_dictionary['Tasks'] = {}

    def get(self, taskID, swimlane_svr):
        task_response = swimlane_svr.swimlane.request(
            'get',
            '/task/{}'.format(taskID),
        )

        task_raw = json.loads(task_response.text)

        return task_raw

    def delete(self, taskID, swimlane_svr):
        taskdelete_response = swimlane_svr.swimlane.request(
            'delete',
            '/task/{}'.format(taskID),
        )

    def update(self, task_obj, task_id, swimlane_svr):

        ## Create Tasks
        update_response = swimlane_svr.swimlane.request(
            'put',
            '/task/{}'.format(task_id),
            json = task_obj
        )

        return json.loads(update_response.text)


    def create(self, task_obj, swimlane_svr):

        ## Create Tasks
        create_response = swimlane_svr.swimlane.request(
            'post',
            '/task/',
            json = task_obj
        )

        create_raw = json.loads(create_response.text)

        ## Add to Menu

        menu_obj = {
            "type": "task",
            "taskId": create_raw['id'],
            "name": create_raw['name'],
            "disabled": False
        }

        if 'applicationId' in task_obj:
            menu_obj["parentId"] = create_raw['applicationId']
            menu_obj["applicationId"] = create_raw['applicationId']

        ## Create Tasks
        menu_response = swimlane_svr.swimlane.request(
            'post',
            '/task/menu',
            json = menu_obj
        )
        swimlane_svr.task_dictionary = swimlane_svr.getTaskDictionary()

        return create_raw

    def removeOrphanTasks(self, target = True, source = False):

        if target:
            for task_name, task_info in self.swimlane_target.task_dictionary.items():
                if 'applicationId' in task_info and self.swimlane_target.validateApp(byID=task_info['applicationId']) == False:
                    self.logger.info('Task {} has an invalid application ID'.format(task_name))
                    self.delete(task_info['id'], self.swimlane_target)

        self.swimlane_target.task_dictionary = self.swimlane_target.getTaskDictionary()

        if source:
            for task_name, task_info in self.swimlane_source.task_dictionary.items():
                if 'applicationId' in task_info and self.swimlane_source.validateApp(byID=task_info['applicationId']) == False:
                    self.logger.info('Task {} has an invalid application ID'.format(task_name))
                    self.delete(task_info['id'], self.swimlane_source)

        self.swimlane_source.task_dictionary = self.swimlane_source.getTaskDictionary()


    def resolveRecordLinks(self, byTaskName = None, byTaskObj = None):
        task_obj = None

        if byTaskName is not None:
            task_id = self.swimlane_source.validateTask(byName=byTaskName)
            task_obj = self.get(task_id, self.swimlane_source)

        if byTaskObj is not None:
            task_obj = byTaskObj

        if task_obj is None:
            raise Exception('Unable to resolve task object')

        self.log_dictionary[task_obj['name']]['fields'] = []

        if 'applicationId' in task_obj:
            source_app_fields = self.swimlane_source.getFieldDictinoary(byID=task_obj['applicationId'])
            target_app_fields = self.swimlane_target.getFieldDictinoary(byName=self.swimlane_source.validateApp(byID=task_obj['applicationId']))

            if 'inputMapping' in task_obj:
                for task_input in task_obj['inputMapping']:
                    if 'type' in task_input:
                        if task_input['type'] == 'record' and 'value' in task_input:
                            source_field_name = self.swimlane_source.validateField(source_app_fields, byID=task_input['value'])
                            if source_field_name != False:
                                target_field_id = self.swimlane_source.validateField(target_app_fields, byName=source_field_name)
                                if target_field_id != False and task_input['value'] != target_field_id:
                                    self.log_dictionary[task_obj['name']]['fields'].append(buildLogDictionary('recordMapping', 'update', source_field_name, 'From: {}, To {}'.format(task_input['value'], target_field_id)))
                                    task_input['value'] = target_field_id

                        elif task_input['type'] == 'asset':
                            source_asset_name = self.swimlane_source.validateAsset(byID=task_input['value']).get('name')
                            if source_asset_name == False:
                                self.logger.critical('Task {} Output mapping asset id {} not found'.format(task_obj['name'], task_input['value']))
                                continue

                            target_asset_id = self.swimlane_target.validateAsset(byName=source_asset_name).get('id')
                            if target_asset_id == False:
                                self.logger.critical('Task {} Output mapping asset name {} not found'.format(task_obj['name'], source_asset_name))
                                continue

                            if target_asset_id != task_input['value']:
                                self.log_dictionary[task_obj['name']]['fields'].append(buildLogDictionary('assetMapping', 'update', source_asset_name, 'From: {}, To {}'.format(task_input['value'], target_asset_id)))
                                task_input['value'] = target_asset_id


            if 'outputs' in task_obj:
                for output in task_obj['outputs']:
                    if isinstance(output, dict) and 'mappings' in output:
                        for output_task in output['mappings']:
                            if 'value' in output_task:
                                source_field_name = self.swimlane_source.validateField(source_app_fields, byID=output_task['value'])
                                if source_field_name != False:
                                    target_field_id = self.swimlane_source.validateField(target_app_fields, byName=source_field_name)
                                    if target_field_id != False and output_task['value'] != target_field_id:
                                        self.log_dictionary[task_obj['name']]['fields'].append(buildLogDictionary('outputs', 'update', source_field_name, 'From: {}, To {}'.format(output_task['value'], target_field_id)))
                                        output_task['value'] = target_field_id

                    elif isinstance(output,dict) and output['type'] == 'referentialTask':
                        source_task_name = self.swimlane_source.validateTask(byID=output['taskId'])
                        target_task_id = self.swimlane_target.validateTask(byName=source_task_name)
                        if target_task_id == False:
                            self.logger.info('Task {} not found in Target'.format(source_task_name))
                            # source_task_name = self.swimlane_source.validateTask(byID = self.layout_obj['taskId'])
                            if source_task_name == False:
                                self.logger.error('FAILED TO FIND IN SOURCE: Orphan Task RAW Data: {}'.format(self.layout_obj['name']))
                                return False
                            else:
                                self.logger.info('Orphan Cross Reference APP: {}'.format(source_task_name))
                                task_raw = self.migrateTask(source_task_name)
                                if 'id' in task_raw:
                                    self.logger.info('Task ID found in target: {}'.format(task_raw['id']))
                                    output['taskId'] = task_raw['id']
                                    self.updates_made = True
                                else:
                                    self.logger.error('FAILED TO FIND: Target Task ID')
                                    return False
                        else:
                            output['taskId'] = target_task_id


        return task_obj


    def migrateTask(self, taskName):
        task_info = self.swimlane_source.validateTask(byName = taskName)
        if task_info != False:
            self.logger.info('Migrating Task {}'.format(taskName))

            full_task = self.get(task_info, self.swimlane_source)
            ##Remove Fields that shouldn't be migrated
            new_task = copy(full_task)
            new_task.pop('id', None)
            new_task.pop('createdByUser', None)
            new_task.pop('modifiedByUser', None)
            new_task.pop('createdDate', None)
            new_task.pop('modifiedDate', None)

            self.log_dictionary[taskName] = {}
            new_task = self.resolveRecordLinks(byTaskObj = new_task)

            if 'applicationId' in new_task:
                source_app_name = self.swimlane_source.validateApp(byID=new_task['applicationId'])
                if source_app_name != False:
                    new_task['applicationId'] = self.swimlane_target.validateApp(byName=source_app_name)
                    if new_task['applicationId'] == False:
                        raise Exception('Unable to find target application {}'.format(source_app_name))
                else:
                    raise Exception('Unable to find source application name for {}'.format(new_task['applicationId']))

            if 'packageDescriptorId' in new_task['action']:
                s_action = self.swimlane_source.action_dictionary[new_task['action']['packageDescriptorId']]
                s_version = s_action['version']
                found = False
                for k, t_act in self.swimlane_target.action_dictionary.iteritems():
                    t_version = t_act.get('version')
                    if t_act['name'] == s_action['name'] and t_version == s_version:
                        new_task['action']['packageDescriptorId'] = t_act['id']
                        new_task['action']['descriptor']['id'] = t_act['id']
                        new_task['action']['descriptor']['packageDescriptor'] = t_act['packageDescriptor']
                        found = True
                        break
                if not found:
                    filename = '{}.swimbundle'.format(s_action['packageDescriptor']['name'])
                    self.logger.info('{} not found. Migrating...'.format(filename))
                    new_bundle = self.copy_swimbundle(s_action['packageDescriptor']['fileId'], filename)
                    self.swimlane_target.action_dictionary = self.swimlane_target.getActionDictionary()
                    for k, t_act in self.swimlane_target.action_dictionary.iteritems():
                        if t_act['name'] == s_action['name'] and t_act['assetDependencyVersion'] == s_action['assetDependencyVersion']:
                            new_task['action']['packageDescriptorId'] = t_act['id']
                            new_task['action']['descriptor']['id'] = t_act['id']
                            new_task['action']['descriptor']['packageDescriptor'] = t_act['packageDescriptor']
                            found = True
                            break
                    if not found:
                        raise Exception('No action found on target named {} with version number {}'.format(s_action['name'],s_action['assetDependencyVersion']))

            if 'assetId' in new_task['action']:
                new_task['action']['assetId'] = self.convert_asset(new_task['action']['assetId'])

            for trigger in new_task['triggers']:
                if 'assetId' in trigger:
                    trigger['assetId'] = self.convert_asset(trigger['assetId'])

            for output in new_task['outputs']:
                if 'assetId' in output:
                    output['assetId'] = self.convert_asset(output['assetId'])

            target_task = self.swimlane_target.validateTask(byName = taskName)
            if target_task != False:
                self.logger.info('Task is on Target - Merge Changes with app ID: {}'.format(target_task))
                new_task['id'] = target_task
                self.log_dictionary[taskName]['action'] = 'update'
                return self.update(new_task, target_task, self.swimlane_target)
            else:
                self.log_dictionary[taskName]['action'] = 'create'
                return self.create(new_task, self.swimlane_target)



        else:
            raise Exception('Failed to find Task {}'.format(taskName))


    def convert_asset(self, src_asset_id):
        src_asset = self.swimlane_source.asset_dictionary_id[src_asset_id]
        try:
            target_asset = self.swimlane_target.asset_dictionary_name[src_asset['name']]
        except:
            raise Exception('No asset on target named {}'.format(src_asset['name']))
        if 'version' in src_asset:
            if target_asset['version'] == src_asset['version']:
                dst_asset_id = target_asset['id']
                return dst_asset_id
            else:
                raise Exception('No asset on target named {} with version {}.'.format(src_asset['name'],
                                                                                      src_asset['version']))
        else:
            dst_asset_id = target_asset['id']
            return dst_asset_id


    def copy_swimbundle(self, file_id, filename):
        stream = StringIO()
        response = self.swimlane_source.swimlane.request('get', 'attachment/download/{}'.format(file_id), stream=True)
        for chunk in response.iter_content(1024):
            stream.write(chunk)
        stream.seek(0)

        upload = self.swimlane_target.swimlane.request('post', "/task/packages",
                                              files={'file': (filename, stream.read())})
        return upload.json()




import types, tasks
from copy import copy
from swimlane_migrator.utils import buildLogDictionary, get_recursively
from sys import exit

class layoutEngine():

    def __init__(self, appName, config):

        self.config = copy(config)

        self.logger = config.pop('logger', None)
        self.swimlane_source = config.pop('swimlane_source', None)
        self.swimlane_target = config.pop('swimlane_target', None)
        self.mode = config.pop('mode', 'Validation')

        if self.logger is None:
            raise  Exception('You must provide a logger object')

        if self.swimlane_source is None:
            raise  Exception('You must provide a swimlane_source object')

        if self.swimlane_target is None:
            raise  Exception('You must provide a swimlane_target object')

        self.logger.info('Layout Engine Running in {} mode'.format(self.mode))

        self.appName = appName

        self.logger.debug('Load Target App Details')
        self.target_app_raw = self.swimlane_target.loadApp(appName=appName)
        self.target_fields = self.swimlane_target.getFieldDictinoary(byName = appName)

        self.logger.debug('Load Source App Details')
        self.source_app_raw = self.swimlane_source.loadApp(appName=appName)
        self.source_fields = self.swimlane_source.getFieldDictinoary(byName = appName)

        self.task_engine = tasks.tasksEngine(config={'logger':self.logger, 'swimlane_source' : self.swimlane_source, 'swimlane_target' : self.swimlane_target})
        self.migrated_tasks = []

        self.log_dictionary = self.swimlane_target.log_dictionary['Layout'] = {}

        self.update_fields = ['html', 'id', 'row', 'col', 'sizex', 'sizey']

    def updateLayout(self):
        if self.mode == 'Live':
            self.logger.info('Remove Orphan Tasks')
            self.task_engine.removeOrphanTasks()

        # # Process Layout
        new_layout = copy(self.source_app_raw['layout'])

        self.log_dictionary['Fields'] = []
        for layout_obj in new_layout:
            self.logger.debug('-------------------------------------')
            layout_obj = self.fix_field_ids(layout_obj, 'fieldId')
            self.logger.debug('-------------------------------------')

        self.target_app_raw['layout'] = new_layout

        if self.mode == 'Live':
            self.swimlane_target.saveApp(self.appName)

    def fix_field_ids(self, search_dict, field):
        for key, value in search_dict.iteritems():
            if key == field:
                source_field_name = self.swimlane_source.validateField(self.source_fields, byID=value)
                if source_field_name == False:
                    self.logger.critical('LAYOUTS->FIX_FIELD_ID : Unable to find source field name: {}'.format(value))
                    continue

                target_field_id = self.swimlane_target.validateField(self.target_fields, byName=source_field_name)
                if target_field_id == False:
                    self.logger.critical('LAYOUTS->FIX_FIELD_ID : Unable to find Target Field Name: {} Field ID: {}'.format(source_field_name, value))
                    target_layout_obj_list = get_recursively(self.target_app_raw['layout'], 'fieldId', [target_field_id])
                else:
                    target_layout_obj_list = []

                ##updateField
                if len(target_layout_obj_list) == 1:
                    # self.logger.critical('##LAYOUT FIELD: {}'.format(source_field_name))
                    modified_fields = []
                    target_layout_obj = target_layout_obj_list[0]
                    for field in self.update_fields:
                        if field in search_dict and field in target_layout_obj:
                            # self.logger.critical('old: {} new: {}'.format(target_layout_obj[field], search_dict[field]))
                            if search_dict[field] != target_layout_obj[field]:
                                # self.logger.critical('Updateing')
                                modified_fields.append(field)
                    # self.logger.critical('Count: {}'.format(len(modified_fields)))
                    if len(modified_fields) > 0:
                        # self.logger.critical('Field {} should be in logger'.format(source_field_name))
                        self.log_dictionary['Fields'].append(buildLogDictionary('fields', 'update', source_field_name, txt = 'Modifed Fields: {}'.format(modified_fields)))
                else:
                    self.log_dictionary['Fields'].append(buildLogDictionary('fields', 'add', source_field_name))

                search_dict[key] = target_field_id


            if '$type' in search_dict and search_dict['$type'] == 'Core.Models.Layouts.IntegrationLayout, Core' and search_dict['id'] not in self.migrated_tasks:
                self.logger.debug('----------------------------------------------')
                layout_obj_build = types.integration(layout_obj = search_dict, config=copy(self.config))
                layout_obj_build.validateTaskID(self.task_engine, self.mode)
                self.migrated_tasks.append(search_dict['id'])
                search_dict = layout_obj_build.layout_obj
                if layout_obj_build.updates_made:
                    self.logger.debug('Modifications Made')
                    #search_dict = layout_obj_build.layout_obj
                self.logger.debug('----------------------------------------------')

            elif isinstance(value, dict):
                search_dict[key] = self.fix_field_ids(value, field)

            elif isinstance(value, list):
                search_dict[key] = []
                for item in value:
                    if isinstance(item, dict):
                        search_dict[key].append(self.fix_field_ids(item, field))
                    else:
                        search_dict[key].append(value)

        return search_dict


import types
from copy import copy
from sys import exit
from swimlane_migrator.utils import buildLogDictionary, find_in_list

class fieldsEngine():

    def __init__(self, appName, config):

        self.config = copy(config)

        self.logger = config.pop('logger', None)
        self.swimlane_source = config.pop('swimlane_source', None)
        self.swimlane_target = config.pop('swimlane_target', None)
        self.mode = config.pop('mode', 'Validation')

        if self.logger is None:
            raise  Exception('You must provide a logger object')

        if self.swimlane_source is None:
            raise  Exception('You must provide a swimlane_source object')

        if self.swimlane_target is None:
            raise  Exception('You must provide a swimlane_target object')

        self.logger.info('Field Engine Running in {} mode'.format(self.mode))

        self.appName = appName

        self.logger.debug('Load Source App Details')
        self.source_app_raw = self.swimlane_source.loadApp(appName=appName)
        self.source_fields = self.swimlane_source.getFieldDictinoary(byName = appName)

        self.logger.debug('Load Target App Details')

        if appName not in self.swimlane_target.apps_dictionary:
            description = ''
            if 'description' in self.source_app_raw:
                description = self.source_app_raw['description']
            if self.mode == 'Live':
                self.swimlane_target.createEmptyApp(appName, acronym = self.source_app_raw['acronym'], id = self.source_app_raw['id'], name = self.source_app_raw['name'], description=description)
            self.logger.info('Create New Application {}'.format(appName))

        self.target_app_raw = self.swimlane_target.loadApp(appName=appName)
        self.target_fields = self.swimlane_target.getFieldDictinoary(byName = appName)

        self.update_field_list = ['required','readOnly','helpTextType', 'layoutId', 'lengthType', 'visualize', 'visualizeMode', 'selectionType', 'controlType', 'inputType', 'name']

        self.migrate_fields_types = ['numeric','valuesList']

        self.migrated_fields = []

        self.temp_field_postfix = 'migration_field_temp'

        self.enable_data_migration = False

        self.log_dictionary = self.swimlane_target.log_dictionary['Fields'] = {}

    def repairTargetFields(self):
        #self.deleteTargetFields(False)
        self.syncTargetFieldInfo()
        self.deleteTargetFields(False)
        self.syncPermissions()

        if self.mode == 'Live':
            self.swimlane_target.saveApp(self.appName)

            if len(self.migrated_fields) > 0 and self.enable_data_migration:
                self.migrateRecordData()

            self.swimlane_target.saveApp(self.appName)

    def deleteTargetFields(self, leaveMigrateFields = True):
        self.log_dictionary['Delete'] = []
        diff_fields = list(set(self.target_fields) - set(self.source_fields))
        check_fields = []
        for df in diff_fields:
            check_fields.append(self.target_fields[df])
        for field in check_fields:

            target_field_info_key = find_in_list(self.target_app_raw['fields'], 'id', field['id'], self.logger)
            if target_field_info_key is None:
                raise Exception('Unable to find field key for: {}'.format(field))

            if self.temp_field_postfix not in field or leaveMigrateFields:
                self.log_dictionary['Delete'].append(buildLogDictionary('field', 'delete', field))
                self.logger.info('Deleting Field {} from target app'.format(field['name']))
                self.target_app_raw['fields'].pop(target_field_info_key)
            else:
                root_field_name = field['name'].replace('_{}'.format(self.temp_field_postfix), '')
                self.migrated_fields.append(root_field_name)

    def syncTargetFieldInfo(self):
        self.log_dictionary['Fields'] = []

        ## Cycle through fields
        for field_obj in self.source_app_raw['fields']:
            self.logger.debug('----------------------------------------------')
            self.logger.debug('Working with Field: {}'.format(field_obj['name']))

            # if field_obj['name'] == 'INC-Did GWO approve closure of ticket?':
            #   self.logger.critical(field_obj)
            '''
            if 'fieldType' in field_obj and field_obj['fieldType'] != 'tracking':
              target_field_info_key = find_in_list(self.target_app_raw['fields'], 'id', field_obj['id'], self.logger)
            else:
              target_field_info_key = find_in_list(self.target_app_raw['fields'], 'name', field_obj['name'], self.logger)
            '''
            target_field_info_key = find_in_list(self.target_app_raw['fields'], 'id', field_obj['id'], self.logger)
            if target_field_info_key is None:
                target_field_info_key = find_in_list(self.target_app_raw['fields'], 'name', field_obj['name'], self.logger)
            if target_field_info_key is None:
                self.logger.info('New Field {} add to field list'.format(field_obj['name']))
                new_field = copy(field_obj)
                self.target_app_raw['fields'].append(new_field)
                current_field_obj = self.target_app_raw['fields'][-1]
                self.log_dictionary['Fields'].append(buildLogDictionary('field', 'add', new_field['name']))

            else:

                self.logger.debug('Current Field')
                current_field_obj = self.target_app_raw['fields'][target_field_info_key]
                ## Validate Field Types Match
                if ('fieldType' in field_obj and 'fieldType' in current_field_obj) and (
                        field_obj['fieldType'] == current_field_obj['fieldType'] or
                        ( field_obj['fieldType'] == 'text' and current_field_obj['fieldType'] in self.migrate_fields_types and self.enable_data_migration)
                ):

                    if field_obj['fieldType'] != current_field_obj['fieldType']:
                        self.log_dictionary['Fields'].append(buildLogDictionary('field', 'migrate', field_obj['name'], 'from {} to {}'.format(field_obj['fieldType'], current_field_obj['fieldType'])))

                        ## Add to migrated list
                        self.migrated_fields.append(field_obj['name'])

                        temp_field_name = '{}_{}'.format(field_obj['name'], self.temp_field_postfix)
                        temp_field_check = find_in_list(self.target_app_raw['fields'], 'name', temp_field_name, self.logger)

                        ## Rename Orginal Field for data migration
                        if temp_field_check is None:
                            current_field_obj['name'] = temp_field_name

                            ##Add New Object
                            new_field = copy(field_obj)
                            self.target_app_raw['fields'].append(new_field)
                            self.logger.critical('Renaming Import field {} to Temp Name {} for migration of data'.format(field_obj['name'], temp_field_name))
                    else:
                        updated_fields = []
                        for update_field in self.update_field_list:
                            if update_field in field_obj:

                                if update_field not in current_field_obj or (current_field_obj[update_field] != field_obj[update_field]):
                                    updated_fields.append(update_field)
                                    self.logger.info('Changing field {} {} from {} to {}'.format(field_obj['name'],update_field,current_field_obj[update_field], field_obj[update_field]))

                                current_field_obj[update_field] = field_obj[update_field]
                            else:

                                current_field_obj.pop(update_field, None)

                        if len(updated_fields) > 0:
                            self.log_dictionary['Fields'].append(buildLogDictionary('field', 'update', field_obj['name'], 'Modified Fields: {}'.format(updated_fields)))

                        if '$type' in current_field_obj:
                            if current_field_obj['$type'] == 'Core.Models.Fields.Reference.ReferenceField, Core':
                                cross_ref_field = types.cross_reference(field_obj = current_field_obj, config=copy(self.config))
                                cross_ref_field.validateTargetApplication()
                                cross_ref_field.validateFieldColumns()
                                if cross_ref_field.updates_made:
                                    current_field_obj = cross_ref_field.field_obj

                            if current_field_obj['$type'] == 'Core.Models.Fields.ValuesListField, Core':
                                value_list_obj = types.valuelist(field_obj=current_field_obj, source_obj=field_obj ,config=copy(self.config))
                                current_field_obj = value_list_obj.migrateValues()
                else:
                    if ('fieldType' in field_obj and 'fieldType' in current_field_obj):
                        self.logger.critical('Unable to migrate Field: {}, Source Field Type: {}  Target Field Type: {}'.format(field_obj['name'], field_obj['fieldType'], current_field_obj['fieldType']))
                    else:
                        self.logger.critical('Unable to migrate Field: {}, Field Type Missing Cannot validate')


            ## Global Modification Fields
            if current_field_obj['$type'] == 'Core.Models.Fields.UserGroupField, Core':
                usergroup_fields = types.usergroups(field_obj = current_field_obj, source_obj=field_obj, config=copy(self.config))
                current_field_obj = usergroup_fields.migrateMembers()
            self.target_fields = self.swimlane_target.getFieldDictinoary(byName = self.appName)
            self.logger.debug('----------------------------------------------')

    def syncPermissions(self):
        self.log_dictionary['Permissions'] = []

        orginal_target_permissions = copy(self.target_app_raw['permissions'])
        self.target_app_raw['permissions'] = {}

        for perm_id, perm in self.source_app_raw['permissions'].items():
            new_perm_dict = None
            target_role_id = None

            if perm_id != '$type':
                if perm['type'] == 'Role':
                    self.logger.info('Update Application Permission for Role: {}'.format(perm['name']))
                    target_role_id = self.swimlane_target.validateRole(byName=perm['name'])

                    if target_role_id == False:
                        self.logger.critical('Failed to find Role: {}, in Target'.format(perm['name']))
                        continue

                    if target_role_id in orginal_target_permissions:
                        self.logger.debug('Role already in target')
                        perm_log = {'group' : 'role', 'action' : 'update', 'name': perm['name'], 'fields' : []}
                    else:
                        self.logger.debug('Add role in target')
                        perm_log = {'group' : 'role', 'action' : 'add', 'name': perm['name'], 'fields' : []}

                    new_perm_dict = copy(perm)
                    new_perm_dict['id'] = target_role_id

                    new_field_perm = {}
                    for field_id, field_perm  in new_perm_dict['fields'].items():
                        if field_id != '$type':

                            source_target_name = self.swimlane_source.validateField(self.source_fields, byID=field_id)
                            if source_target_name == False:
                                self.logger.critical('Failed to assign Field Permissions, ID {} not found in source'.format(field_id))
                                continue

                            target_field_id = self.swimlane_target.validateField(self.target_fields, byName=source_target_name)
                            if target_field_id == False:
                                self.logger.critical('Failed to assign Field Permissions, Name {} not found in target'.format(source_target_name))
                                continue

                            if target_role_id in orginal_target_permissions:
                                if target_field_id in orginal_target_permissions[target_role_id]['fields']:
                                    if field_perm != orginal_target_permissions[target_role_id]['fields'][target_field_id]:
                                        self.logger.info('Updating Permission for Field: {}, Old: {}  New: {}'.format(source_target_name, orginal_target_permissions[target_role_id]['fields'][target_field_id], field_perm))
                                        perm_log['fields'].append(buildLogDictionary('field', 'update', source_target_name, 'Old: {}, New {}'.format(orginal_target_permissions[target_role_id]['fields'][target_field_id], field_perm)))
                                else:
                                    self.logger.info('Adding Permission for Field: {}, Set to: {}'.format(source_target_name, field_perm))
                                    perm_log['fields'].append(buildLogDictionary('field', 'add', source_target_name, 'New {}'.format(field_perm)))

                                new_field_perm[target_field_id] = field_perm
                            else:
                                #self.logger.critical('Target Role ID not found {}'.format(target_role_id))
                                self.logger.info('Adding Permission for Field: {}, Set to: {}'.format(source_target_name, field_perm))
                                perm_log['fields'].append(buildLogDictionary('field', 'add', source_target_name, 'New {}'.format(field_perm)))
                                if len(perm_log['fields']) > 0:
                                    self.log_dictionary['Permissions'].append(perm_log)
                                new_field_perm[target_field_id] = field_perm
                                new_perm_dict['fields'] = new_field_perm
                    if len(perm_log['fields']) > 0:
                        self.log_dictionary['Permissions'].append(perm_log)

                    new_perm_dict['fields'] = new_field_perm

                else:
                    raise Exception('Unhandled Permission Type {}'.format(perm['type']))

            if new_perm_dict is not None and target_role_id is not None:
                self.target_app_raw['permissions'][target_role_id] = new_perm_dict

    def migrateRecordData(self):
        current_app = self.swimlane_target.swimlane.apps.get(name=self.appName)
        search_results = current_app.records.search(('Tracking Id', 'doesNotEqual', ''))
        for record in search_results:
            for field_name in self.migrated_fields:
                self.logger.critical(field_name)
                temp_field_name = '{}_{}'.format(field_name, self.temp_field_postfix)
                if record[temp_field_name] is not None:
                    record[field_name] = str(record[temp_field_name])
                self.logger.critical(record[temp_field_name])
            record.save()

        self.deleteTargetFields(leaveMigrateFields=False)
        self.swimlane_target.saveApp(self.appName)



from base.fields import fields

class valuelist(fields):

    def __init__(self, field_obj, source_obj, config ):
        super(valuelist, self).__init__(field_obj, config)
        self.source_object = source_obj

    def migrateValues(self):
        self.logger.debug(self.field_obj)
        target_value_list = self.field_obj['values']
        source_value_list = self.source_object['values']

        new_target_value_list = source_value_list

        ## Add New
        for target_value in new_target_value_list:
            target_value_list_key = self.find_in_list(target_value_list, 'name', target_value['name'])
            ## Add If not found
            if target_value_list_key is not None:
                target_value['id'] = target_value_list[target_value_list_key]['id']

        self.field_obj['values'] = new_target_value_list


        return self.field_obj


    def find_in_list(self, lst, key, value):
        for i, dic in enumerate(lst):
            self.logger.debug('Key: {} Value: {} Check: {}'.format(key, dic[key], value))
            if key in dic and dic[key] == value:
                self.logger.debug('####FOUND####')
                return i
        return None


from base.fields import fields
from copy import copy
import sys

class usergroups(fields):

    def __init__(self, field_obj, source_obj, config ):
        super(usergroups, self).__init__(field_obj, config)
        self.source_object = source_obj

    def migrateMembers(self):
        source_members = self.source_object['members']
        new_members_list = []

        for member in self.source_object['members']:
            new_member = copy(member)

            if member['itemType'] == 'group':
                try:
                    self.logger.info('Looking up Group: {}'.format(member))
                    sw_group = self.swimlane_target.swimlane.groups.get(name=member['name'])
                    new_member['id'] = sw_group.id
                    new_members_list.append(new_member)
                except ValueError as e:
                    self.logger.error('Group: {} not found removing'.format(member['name']))


            elif member['itemType'] == 'user':
                try:
                    self.logger.info('Looking up User: {}'.format(member))
                    sw_user = self.swimlane_target.swimlane.users.get(display_name=member['name'])
                    new_member['id'] = sw_user.id
                    new_members_list.append(new_member)
                except ValueError as e:
                    self.logger.error('User: {} not found removing'.format(member['name']))

        self.field_obj['members'] = new_members_list

        return self.field_obj


from swimlane import Swimlane
import json
import uuid


class swimlane_server(object):

    def __init__(self, sw_url, sw_api_user, sw_api_password, logger):
        self.logger = logger

        self.logger.info('Connect to Source: {}'.format(sw_url))
        self.swimlane = Swimlane(sw_url, sw_api_user, sw_api_password, verify_ssl=False)

        self.logger.info('Building App Dictionary')
        self.apps_dictionary = self.getApplicationDictionary()
        self.logger.info('Building Task Dictionary')
        self.task_dictionary = self.getTaskDictionary()
        self.roles_dictionary = self.getRoleDictionary()
        self.asset_dictionary_name,self.asset_dictionary_id = self.getAssetDictionary()
        self.action_dictionary = self.getActionDictionary()

        self.log_dictionary = {}

        self.apps_objects = {}

    def updateWorkflow(self, AppName, workflow):
        workflow_response = self.swimlane.request(
            'put',
            '/workflow/{}'.format(workflow['id']),
            json=workflow
        )

        workflow = json.loads(workflow_response.text)

    def saveApp(self, AppName):
        app_obj = self.loadApp(appName=AppName)
        app_response = self.swimlane.request(
            'put',
            '/app/{}'.format(self.apps_dictionary[AppName]['id']),
            json = app_obj
        )

        self.apps_objects[self.apps_dictionary[AppName]['id']] = json.loads(app_response.text)

    def createEmptyApp(self, appName, **kwargs):

        app_obj = {
            "$type": "Core.Models.Application.Application, Core",
            "acronym": kwargs.pop('acronym', 'FEA'),
            "layout": [],
            "fields": [],
            "maxTrackingId": 0,
            "workspaces": [],
            "createWorkspace": False,
            "timeTrackingEnabled": False,
            "permissions": {
                "$type": "Core.Models.Security.PermissionMatrix, Core"
            },
            "id": kwargs.pop('id', str(uuid.uuid1()).replace('-', '')),
            "name": kwargs.pop('name', 'Fake Empty Application'),
            "description" : kwargs.pop('description', ''),
            "disabled": False
        }

        app_response = self.swimlane.request(
            'post',
            '/app',
            json = app_obj
        )

        raw_app_obj = json.loads(app_response.text)
        self.apps_objects[raw_app_obj['id']] = raw_app_obj
        self.apps_dictionary = self.getApplicationDictionary()

    def getAssetDictionary(self):
        asset_response = self.swimlane.request(
            'get',
            '/asset',
        )
        asset_raw = json.loads(asset_response.text)

        asset_dictionary_name = {}
        asset_dictionary_id = {}
        for asset in asset_raw:
            asset_dictionary_name[asset['name']] = asset
            asset_dictionary_id[asset['id']] = asset

        return asset_dictionary_name,asset_dictionary_id

    def getActionDictionary(self):
        action_response = self.swimlane.request('get','task/actions').json()

        action_dictionary = {}
        for action in action_response:
            action_dictionary[action['id']] = action

        return action_dictionary

    def getApplicationDictionary(self):
        apps_response = self.swimlane.request(
            'get',
            '/app/light',
        )
        apps_raw = json.loads(apps_response.text)

        apps_dictionary = {}
        for app in apps_raw:
            apps_dictionary[app['name']] = app

        return apps_dictionary

    def getRoleDictionary(self):
        roles_response = self.swimlane.request(
            'get',
            '/roles/light',
        )
        roles_raw = json.loads(roles_response.text)

        roles_dictionary = {}
        for role in roles_raw:
            roles_dictionary[role['name']] = role

        return roles_dictionary

    def getTaskDictionary(self):
        task_response = self.swimlane.request(
            'get',
            '/task/light',
        )
        task_raw = json.loads(task_response.text)

        tasks_dictionary = {}
        for task in task_raw:
            tasks_dictionary[task['name']] = task

        return tasks_dictionary

    def getFieldDictinoary(self, byID = None, byName = None, forceReload = False):
        if byID is not None:
            app_obj = self.loadApp(appID=byID, fieldsOnly = True, forceReload = forceReload)
        else:
            app_obj = self.loadApp(appName=byName, fieldsOnly = True, forceReload = forceReload)

        field_dictionary = {}
        for field_obj in app_obj:
            if 'id' in field_obj and 'name' in field_obj:
                field_dictionary[field_obj['name']] = field_obj

        return field_dictionary

    def loadWorkflow(self, appID=None, appName=None):
        if appName is not None:
            if appName in self.apps_dictionary:
                appID = self.apps_dictionary[appName]['id']
            else:
                #raise Exception('Unable to find AppID for {}'.format(appName))
                return {}


        workflow_response = self.swimlane.request(
            'get',
            '/workflow/{}'.format(appID),
        )

        return json.loads(workflow_response.text)

    def loadApp(self, appID=None, appName=None, forceReload = False, **kwargs):
        if appName is not None:
            if appName in self.apps_dictionary:
                appID = self.apps_dictionary[appName]['id']
            else:
                #raise Exception('Unable to find AppID for {}'.format(appName))
                return {"fields": [], "permissions": [], "layout": []}

        if forceReload or appID not in self.apps_objects:
            appdetail_response = self.swimlane.request(
                'get',
                '/app/{}'.format(appID),
            )

            self.apps_objects[appID] = json.loads(appdetail_response.text)

        if kwargs.pop('fieldsOnly', False):
            return self.apps_objects[appID]['fields']

        if kwargs.pop('layoutOnly', False):
            return self.apps_objects[appID]['layout']

        return self.apps_objects[appID]

    def validateAsset(self, byID = None, byName = None):
        if byID is not None:
            return self.asset_dictionary_id.get(byID, dict())

        if byName is not None:
            return self.asset_dictionary_name.get(byName,dict())

    def validateApp(self, byID = None, byName = None):
        if byID is not None:
            return self.lookupLabel(self.apps_dictionary, byValue=byID)

        if byName is not None:
            return self.lookupLabel(self.apps_dictionary, byKey=byName)

    def validateField(self, field_dictionary, byID = None, byName = None):
        if byID is not None:
            return self.lookupLabel(field_dictionary, byValue=byID)

        if byName is not None:
            return self.lookupLabel(field_dictionary, byKey=byName)

    def validateTask(self, byID = None, byName = None):
        if byID is not None:
            return self.lookupLabel(self.task_dictionary, byValue=byID)

        if byName is not None:
            return self.lookupLabel(self.task_dictionary, byKey=byName)

    def validateRole(self, byID = None, byName = None):
        if byID is not None:
            return self.lookupLabel(self.roles_dictionary, byValue=byID)

        if byName is not None:
            return self.lookupLabel(self.roles_dictionary, byKey=byName)

    def lookupLabel(self, lookup_dictionary, byValue=None, byKey=None):

        if byValue is not None:
            for key, value in lookup_dictionary.items():
                if byValue == value['id']:
                    return key

        if byKey is not None:
            if byKey in lookup_dictionary:
                return lookup_dictionary[byKey]['id']

        return False



import json, sys
from copy import copy
from swimlane_migrator.utils import lookupLabel
from .reports import report

class  dashboard(object):

    def __init__(self, config):

        self.logger = config.pop('logger', None)
        self.swimlane_source = config.pop('swimlane_source', None)
        self.swimlane_target = config.pop('swimlane_target', None)

        if self.swimlane_source is None:
            raise  Exception('You must provide a swimlane_source object')

        if self.swimlane_target is None:
            raise  Exception('You must provide a swimlane_target object')

        self.remove_keys = ['createdDate', 'modifiedDate', 'createdByUser', 'modifiedByUser']
        self.mode = config['mode']
        self.dashboard_dictionary_source = self.getDashboardDictionary(self.swimlane_source)
        self.dashboard_dictionary_target = self.getDashboardDictionary(self.swimlane_target)

        self.report_engine = report(config={'logger':self.logger, 'swimlane_source' : self.swimlane_source, 'swimlane_target' : self.swimlane_target, 'mode': self.mode})


    def getDashboards(self, swimlane_server):
        dashboard_response = swimlane_server.swimlane.request(
            'get',
            '/dashboard'
        )

        return json.loads(dashboard_response.text)

    def getDashboard(self, swimlane_server, dashboard_id):
        dashboard_response = swimlane_server.swimlane.request(
            'get',
            '/dashboard/{}'.format(dashboard_id)
        )
        if dashboard_response.text != '':
            return json.loads(dashboard_response.text)
        return None


    def deleteDashboard(self, swimlane_server, dashboard_id):
        workspace_response = swimlane_server.swimlane.request(
            'delete',
            '/dashboard/{}'.format(dashboard_id)
        )

    def addDashboard(self, swimlane_server, dashboard):
        dashboard_response = swimlane_server.swimlane.request(
            'post',
            '/dashboard',
            json = dashboard
        )

        return json.loads(dashboard_response.text)

    def updateDashboard(self, swimlane_server, dashboard_id, dashboard):
        dashboard_response = swimlane_server.swimlane.request(
            'put',
            '/dashboard/{}'.format(dashboard_id),
            json = dashboard
        )

        return json.loads(dashboard_response.text)

    def getDashboardDictionary(self, swimlane_server):
        dashboard_response = self.getDashboards(swimlane_server)

        dashboard_dictionary = {}
        for dashboard in dashboard_response:
            dashboard_dictionary[dashboard['name']] = dashboard

        return dashboard_dictionary

    def validateDashboard(self, source, byID = None, byName = None):
        lookup_dictionary = None

        if source == 'target':
            lookup_dictionary = self.dashboard_dictionary_target

        if source == 'source':
            lookup_dictionary = self.dashboard_dictionary_source

        if lookup_dictionary is None:
            raise Exception('Invalid source "{}" provided, valid options "target" or "source"'.format(source))

        if byID is not None:
            return lookupLabel(lookup_dictionary, byValue=byID)

        if byName is not None:
            return lookupLabel(lookup_dictionary, byKey=byName)

    def removeOrphanDashboards(self, swimlane_server):
        dashboards = self.getDashboards(swimlane_server)
        for dashboard in dashboards:
            self.logger.info('Working on Dashboard: {}, which has {} workspaces'.format(dashboard['name'], len(dashboard['workspaces'])))
            if 'workspaces' in dashboard and len(dashboard['workspaces']) == 0:
                self.logger.error('Removing Dashboard {}'.format(dashboard['name']))
                self.deleteDashboard(swimlane_server, dashboard['id'])

    def fixItems(self, items):

        for item in items:
            if 'reportId' in item:
                source_report_obj = self.report_engine.getReportByID(self.swimlane_source, item['reportId'])
                if 'id' not in source_report_obj:
                    self.logger.critical('Dashboard Card Report ID {} could not be found in source'.format(item['reportId']))
                    continue

                source_app_name = self.swimlane_source.validateApp(byID=source_report_obj['applicationIds'][0])
                if source_app_name == False:
                    self.logger.critical('Dashboard Card App ID {} could not be found in source'.format(source_report_obj['applicationIds'][0]))
                    continue

                target_app_id = self.swimlane_target.validateApp(byName=source_app_name)
                if target_app_id == False:
                    self.logger.critical('Dashboard Card App Name "{}" could not be found in target'.format(source_app_name))
                    continue

                target_report_dictionary = self.report_engine.getReportDictionary(self.swimlane_target, target_app_id)
                target_report_id = lookupLabel(target_report_dictionary, byKey=source_report_obj['name'])
                if target_report_id == False:
                    self.logger.critical('Dashboard Card Report Name "{}" could not be found in target'.format(source_report_obj['name']))
                    continue

                item['reportId'] = target_report_id

        return items


    def migrateDashboards(self, mode,workspace_obj, target_workspace_id, target_workspace_list):


        for dashboard_id in workspace_obj['dashboards']:
            source_dashboard = self.getDashboard(self.swimlane_source, dashboard_id)

            if source_dashboard is None:
                self.logger.critical("Failed to find Dashboard ID {} in source".format(dashboard_id))
                continue

            target_dashboard_id = self.validateDashboard('target', byName=source_dashboard['name'])

            if target_dashboard_id == False:
                new_dashboard = copy(source_dashboard)
                new_dashboard['workspaces'] = [target_workspace_id]

                for key in self.remove_keys:
                    new_dashboard.pop(key, None)

                new_dashboard['items'] = self.fixItems(new_dashboard['items'])
                if mode == 'Live':
                    self.addDashboard(self.swimlane_target, new_dashboard)
                    self.dashboard_dictionary_target = self.getDashboardDictionary(self.swimlane_target)
            else:
                update_dashboard = self.getDashboard(self.swimlane_target, target_dashboard_id)

                for key, workspace_id in enumerate(update_dashboard['workspaces']):
                    if workspace_id not in target_workspace_list:
                        update_dashboard['workspaces'].pop(key)

                if target_workspace_id not in update_dashboard['workspaces']:
                    update_dashboard['workspaces'].append(target_workspace_id)

                update_dashboard['items'] = self.fixItems(source_dashboard['items'])
                if mode == 'Live':
                    self.updateDashboard(self.swimlane_target, target_dashboard_id, update_dashboard)
                    self.dashboard_dictionary_target = self.getDashboardDictionary(self.swimlane_target)



import json, sys

class groups(object):

    def __init__(self, config):

        self.logger = config.pop('logger', None)
        self.swimlane_source = config.pop('swimlane_source', None)
        self.swimlane_target = config.pop('swimlane_target', None)

        if self.swimlane_source is None:
            raise  Exception('You must provide a swimlane_source object')

        if self.swimlane_target is None:
            raise  Exception('You must provide a swimlane_target object')

        self.remove_keys = ['createdDate', 'modifiedDate', 'createdByUser', 'modifiedByUser']
        self.group_exclustions = ['Everyone']
        self.mode = config['mode']

    def getGroups(self, swimlane_server):
        groups_response = swimlane_server.swimlane.request(
            'get',
            '/groups'
        )

        return json.loads(groups_response.text)


    def addGroup(self, swimlane_server, group):
        groups_response = swimlane_server.swimlane.request(
            'post',
            '/groups',
            json = group
        )

        return json.loads(groups_response.text)

    def updateGroup(self, swimlane_server, group_id, group):
        groups_response = swimlane_server.swimlane.request(
            'put',
            '/groups/{}'.format(group_id),
            json = group
        )

        return json.loads(groups_response.text)

    def lookupGroup(self, swimlane_server, name):
        groups_response = swimlane_server.swimlane.request(
            'get',
            '/groups/lookup?name={}'.format(name)
        )

        return json.loads(groups_response.text)

    def migrateGroups(self):
        sw_source_groups_raw = self.getGroups(self.swimlane_source)
        sw_target_groups_raw = self.getGroups(self.swimlane_target)

        for group in sw_source_groups_raw['groups']:
            if group['name'] not in self.group_exclustions:
                self.logger.info('Working with Group: {}'.format(group['name']))

                ## Process Users
                for user_members in group['users']:
                    source_user_obj = self.swimlane_source.swimlane.users.get(id=user_members['id'])
                    try:
                        target_user_obj = self.swimlane_target.swimlane.users.get(display_name=source_user_obj.display_name)
                        user_members['id'] = target_user_obj.id
                        user_members['name'] = target_user_obj.username
                    except ValueError as e:
                        ## Unable to find user - remove user association
                        self.logger.warn('User: {} not found removing from Group: {}'.format(source_user_obj.display_name, group['name']))
                        if self.mode == 'Live':
                            group['users'].remove(user_members)


                for group_members in group['groups']:
                    source_group_obj = self.swimlane_source.swimlane.groups.get(id=group_members['id'])
                    try:
                        target_group_obj = self.swimlane_target.swimlane.groups.get(name=source_group_obj.name)
                        group_members['id'] = target_group_obj.id
                        group_members['name'] = target_group_obj.name
                    except ValueError as e:
                        ## Unable to find group attempt to add
                        self.logger.warn('Group: {} not found attempting to add'.format(source_user_obj.name))
                        new_group = {}
                        new_group['id'] = source_group_obj.id
                        new_group['description'] = source_group_obj.description
                        new_group['name'] = source_group_obj.name
                        new_group['disabled'] =  source_group_obj.disabled
                        if self.mode == 'Live':
                            new_group = addGroup(self.swimlane_target, new_group)
                            sw_target_groups_raw['groups'].append(new_group)

                target_group_key = self.find_in_list(sw_target_groups_raw['groups'], 'name', group['name'])
                self.logger.info('Target Key: {}'.format(target_group_key))
                if target_group_key is None:
                    ## New Group
                    for key in self.remove_keys:
                        if key in group:
                            group.pop(key)

                    ## Had to wrap in try it seems that some groups don't get returned when you ask for the groups
                    try:
                        if self.mode == 'Live':
                            add_group = self.addGroup(self.swimlane_target, group)
                            sw_target_groups_raw['groups'].append(add_group)
                    except Exception as e:
                        update_group_obj = self.swimlane_target.swimlane.groups.get(name=group['name'])
                        update_group = update_group_obj._raw
                        update_group['users'] = group['users']
                        if self.mode == 'Live':
                            self.updateGroup(self.swimlane_target, update_group['id'], update_group)
                            sw_target_groups_raw['groups'].append(update_group)
                else:
                    update_group = sw_target_groups_raw['groups'][target_group_key]
                    update_group['users'] = group['users']
                    update_group['groups'] = group['groups']
                    if self.mode == 'Live':
                        self.updateGroup(self.swimlane_target, update_group['id'], update_group)

    def find_in_list(self, lst, key, value):
        for i, dic in enumerate(lst):
            if key in dic and dic[key] == value:
                return i
        return None






def get_recursively(search_dict, field, search_list):
    class Namespace(object):
        pass
    ns = Namespace()
    ns.results = []

    def inner(search_dict, field, search_list):
        for key, value in search_dict.iteritems():
            if key == field and value in search_list:
                ns.results.append(search_dict)

            elif isinstance(value, dict):
                inner(value, field, search_list)

            elif isinstance(value, list):
                for item in value:
                    if isinstance(item, dict):
                        inner(item, field, search_list)

    if not isinstance(search_dict, dict):
        search_dict = {'data' : search_dict}
    inner(search_dict, field, search_list)
    return ns.results


def find_in_list(lst, key, value, logger):
    for i, dic in enumerate(lst):
        logger.debug('Key: {} Value: {} Check: {}'.format(key, dic[key], value))
        if key in dic and dic[key] == value:
            logger.debug('####FOUND####')
            return i
    return None


def updateLayoutPosition(source_obj, target_obj, logger):
    postion_layout_fields = ['row', 'col', 'sizex', 'sizey']
    for field in postion_layout_fields:
        if field in source_obj and field in target_obj:
            logger.info('Source {}: {} --- Target {}: {}'.format(field, source_obj[field], field, target_obj[field]))
            target_obj[field] = source_obj[field]

    return target_obj


def buildLogDictionary(group, action, name, txt = ''):
    return  {
      'group' : group,
      'action' : action,
      'name' : name,
      'txt' : txt
    }

def lookupLabel(lookup_dictionary, byValue=None, byKey=None):

    if byValue is not None:
        for key, value in lookup_dictionary.items():
            if byValue == value['id']:
                return key

    if byKey is not None:
        if byKey in lookup_dictionary:
            return lookup_dictionary[byKey]['id']

    return False


from setuptools import setup, find_packages

def parse_requirements(requirement_file):
    with open(requirement_file) as f:
        return f.readlines()


with open('.\\README.md') as f:
    long_description = f.read()

setup(
    name='swimlane_migrator',
    version='0.4.2',
    packages=find_packages(),
    url='https://swimlane.com',
    license='',
    author='Swimlane',
    author_email='info@swimlane.com',
    description='Swimlane Content Migration Package',
    long_description=long_description,
    long_description_content_type='text/markdown',
    install_requires=parse_requirements('.\\requirements.txt')
)



from base.layouts import layouts

class integration(layouts):

    def validateTaskID(self, taskEngine, mode):
        if 'taskId' in self.layout_obj:
            source_task_name = self.swimlane_source.validateTask(byID=self.layout_obj['taskId'])
            target_task_id = self.swimlane_target.validateTask(byName= source_task_name)
            if target_task_id == False:
                self.logger.info('Task {} not found in Target'.format(source_task_name))
                #source_task_name = self.swimlane_source.validateTask(byID = self.layout_obj['taskId'])
                if source_task_name == False:
                    self.logger.error('FAILED TO FIND IN SOURCE: Orphan Task RAW Data: {}'.format(self.layout_obj['name']))
                    return False
                else:
                    self.logger.info('Orphan Cross Reference APP: {}'.format(source_task_name))
                    if mode == 'Live':
                        task_raw = taskEngine.migrateTask(source_task_name)
                    else:
                        task_raw = {}
                    # target_task_id = self.swimlane_target.validateTask(byName = source_task_name)
                    if 'id' in task_raw:
                        self.logger.info('Task ID found in target: {}'.format(task_raw['id']))
                        self.layout_obj['taskId'] = task_raw['id']
                        self.updates_made = True
                    else:
                        self.logger.error('FAILED TO FIND: Target Task ID')
                        return False


            else:
                self.logger.debug('Integration Task {} is valid'.format(source_task_name))
                self.layout_obj['taskId'] = target_task_id
                self.updates_made = True
        else:
            self.logger.critical('Task Button {} has no taskID'.format(self.layout_obj['name']))



from copy import copy
from sys import exit
from swimlane_migrator.utils import find_in_list
from swimlane_migrator.layouts.tasks import tasksEngine
import re

class workflowEngine():

    def __init__(self, appName, config):

        self.logger = config.pop('logger', None)
        self.swimlane_source = config.pop('swimlane_source', None)
        self.swimlane_target = config.pop('swimlane_target', None)

        if self.logger is None:
            raise  Exception('You must provide a logger object')

        if self.swimlane_source is None:
            raise  Exception('You must provide a swimlane_source object')

        if self.swimlane_target is None:
            raise  Exception('You must provide a swimlane_target object')

        self.appName = appName
        self.mode = config['mode']

        self.logger.info('Workflow Engine is running in {} mode.'.format(self.mode))

        self.logger.info('Load Target App Details')
        self.target_app_raw = self.swimlane_target.loadApp(appName=appName)
        self.target_fields = self.swimlane_target.getFieldDictinoary(byName = appName)

        self.logger.info('Load Source App Details')
        self.source_app_raw = self.swimlane_source.loadApp(appName=appName)
        self.source_fields = self.swimlane_source.getFieldDictinoary(byName = appName, forceReload = True)

        self.source_workflow = self.swimlane_source.loadWorkflow(appName=appName)
        self.target_workflow = self.swimlane_target.loadWorkflow(appName=appName)

        self.task_engine = tasksEngine(config={'logger': self.logger, 'swimlane_source': self.swimlane_source, 'swimlane_target': self.swimlane_target})

        self.current_workflow_label = ''
        self.migrated_ids = []

    def migrateWorkflow(self):
        new_workflow = copy(self.source_workflow)

        for stage in new_workflow['stages']:
            stage = self.fix_field_ids(stage, 'fieldId')

        self.target_workflow['stages'] = new_workflow['stages']
        if self.mode == 'Live':
            try:
                self.swimlane_target.updateWorkflow(AppName=self.appName, workflow=self.target_workflow)
            except Exception as e:
                raise Exception(e)


    def fixReadWriteFields(self, search_dict):
        new_fieldState = {}
        new_fieldState['$type'] = search_dict['fieldStates']['$type']

        self.logger.debug('Working on Read/Write Acition Named: {}'.format(search_dict['name']))
        self.logger.debug(search_dict['fieldStates'])
        for field_id, field_perm in search_dict['fieldStates'].items():
            if field_id != '$type':
                source_field_name = self.swimlane_source.validateField(self.source_fields, byID=field_id)
                if source_field_name == False:
                    self.logger.error('Action: {} with Field ID {}, not found in source.'.format(search_dict['name'], field_id))
                    continue

                target_field_id = self.swimlane_target.validateField(self.target_fields, byName=source_field_name)
                if target_field_id == False:
                    self.logger.critical('Read Write Field Name {}, not found in target.'.format(source_field_name))
                    continue

                self.logger.info('Updating Field Permissions for {} to {}'.format(source_field_name, field_perm))
                new_fieldState[target_field_id] = field_perm

        search_dict['fieldStates'] = new_fieldState
        return search_dict

    def fixFieldSetValues(self, search_dict):
        current_value = search_dict['value']
        if isinstance(current_value, unicode):
            matches = re.findall('{{(.*?)}}', current_value, re.DOTALL)
            for field_id in matches:
                self.logger.debug('Lookup Field ID: {}'.format(field_id))
                source_field_name = self.swimlane_source.validateField(self.source_fields, byID=field_id)
                if source_field_name == False:
                    self.logger.error('FieldSetValue: {} with Field ID {}, not found in source.'.format(search_dict['name'], field_id))
                    continue

                self.logger.debug('Lookup Field Name: {}'.format(source_field_name))
                target_field_id = self.swimlane_target.validateField(self.target_fields, byName=source_field_name)
                if target_field_id == False:
                    self.logger.critical('FieldSetValue: Field Name {}, not found in target.'.format(source_field_name))
                    continue

                current_value = current_value.replace(field_id, target_field_id)

            search_dict['value'] = current_value

        return search_dict

    def fixTaskIdValues(self, search_dict):
        current_id = search_dict['taskId']
        for task, value in self.swimlane_source.task_dictionary.iteritems():
            if value['id'] == current_id:
                task_name = value['name']
                self.logger.debug('Converting Task Id for: {}'.format(task_name))
                found = False
                for task, value in self.swimlane_target.task_dictionary.iteritems():
                    if value['name'] == task_name:
                        new_id = value['id']
                        search_dict['taskId'] = new_id
                        found = True
                        break
                if not found:
                    if self.mode == "Live":
                        new_task = self.task_engine.migrateTask(task_name)
                        new_id = new_task['id']
                        search_dict['taskId'] = new_id
                break
        return search_dict

    def fixConditionValue(self, search_dict):
        if 'fieldId' in search_dict and 'value' in search_dict:
            source_field_name = self.swimlane_source.validateField(self.source_fields, byID=search_dict['fieldId'])
            if source_field_name == False:
                self.logger.critical('{} ConditionValue: Field ID {}, not found in source. Raw Data:  {}'.format(self.current_workflow_label, search_dict['fieldId'], search_dict))
                return search_dict

            target_field_id = self.swimlane_target.validateField(self.target_fields, byName=source_field_name)
            if target_field_id == False:
                self.logger.critical('{} ConditionValue: Field Name {}, not found in target.'.format(self.current_workflow_label, source_field_name))
                return search_dict

            ## Validate for Value List
            if self.source_fields[source_field_name]['$type'] == 'Core.Models.Fields.ValuesListField, Core':
                try:
                    source_value_obj_key = find_in_list(self.source_fields[source_field_name]['values'], 'id', search_dict['value'], self.logger)
                    if source_value_obj_key is None:
                        self.logger.critical('"{}" ConditionValue: Field Value ID "{}" not found in Field "{}" with ID "{}" in source value List'.format(self.current_workflow_label, search_dict['value'], source_field_name, search_dict['fieldId']))
                        return search_dict

                    source_value_obj = self.source_fields[source_field_name]['values'][source_value_obj_key]
                    target_value_obj_key = find_in_list(self.target_fields[source_field_name]['values'], 'name', source_value_obj['name'], self.logger)
                    if target_value_obj_key is None:
                        self.logger.critical('{} ConditionValue: Field Value {}, not found in target value List.'.format(self.current_workflow_label, source_value_obj['name']))
                        return search_dict

                    search_dict['value'] = self.target_fields[source_field_name]['values'][target_value_obj_key]['id']
                except Exception as e:
                    raise Exception('Error in Fix Condition: {}, Raw Dict: {}'.format(e, search_dict))

        return search_dict


    def fix_field_ids(self, search_dict, field):
        remove_keys = []


        if '$type' in search_dict:
            if search_dict['$type'] == 'Core.Models.Workflow.Stage, Core' or search_dict['$type'] == 'Core.Models.Workflow.Actions.FieldSetAction, Core':
                self.current_workflow_label = search_dict['name']

            if search_dict['$type'] == 'Core.Models.Workflow.Condition, Core':
                search_dict = self.fixConditionValue(search_dict)

            elif search_dict['$type'] == 'Core.Models.Workflow.Actions.FieldStateAction, Core':
                if 'fieldStates' in search_dict and search_dict['id'] not in self.migrated_ids:
                    search_dict = self.fixReadWriteFields(search_dict)
                    self.migrated_ids.append(search_dict['id'])

            elif search_dict['$type'] == 'Core.Models.Workflow.Actions.FieldSetAction, Core':
                if 'value' in search_dict and search_dict['id'] not in self.migrated_ids:
                    search_dict = self.fixFieldSetValues(search_dict)
                    if 'id' in search_dict:
                        self.migrated_ids.append(search_dict['id'])

            elif search_dict['$type'] == 'Core.Models.Workflow.Actions.IntegrationAction, Core':
                if 'taskId' in search_dict and search_dict['id'] not in self.migrated_ids:
                    search_dict = self.fixTaskIdValues(search_dict)
                    if 'taskId' in search_dict:
                        self.migrated_ids.append(search_dict['taskId'])

        for key, value in search_dict.iteritems():
            if key == field:

                source_field_name = self.swimlane_source.validateField(self.source_fields, byID=value)
                if source_field_name == False:
                    if 'name' in search_dict:
                        label = search_dict['name']
                    else:
                        label = search_dict

                    self.logger.critical('Failed to Migrate Workflow Object {}: {}, source Field ID {} Not Found.'.format(self.current_workflow_label, label, value))
                else:
                    target_field_id = self.swimlane_target.validateField(self.target_fields, byName=source_field_name)
                    if target_field_id == False:
                        self.logger.critical('Failed to Migrate Workflow Object: {}, target Field Not Found.'.format(source_field_name))
                    else:
                        if 'name' in search_dict:
                            label = search_dict['name']
                        else:
                            label = search_dict
                        self.logger.debug('Updated Field : {}'.format(label))
                        search_dict[key] = target_field_id


            elif isinstance(value, dict):
                search_dict[key] = self.fix_field_ids(value, field)

            elif isinstance(value, list):
                search_dict[key] = []
                for item in value:
                    if isinstance(item, dict):
                        search_dict[key].append(self.fix_field_ids(item, field))
                    else:
                        search_dict[key] = value

        for key in remove_keys:
            search_dict.pop(key)

        return search_dict


from .dashboards import dashboard
from .reports import report
from swimlane_migrator.utils import find_in_list
import json
from copy import copy
from sys import exit


class workspace(object):

    def __init__(self, config):

        self.logger = config.pop('logger', None)
        self.swimlane_source = config.pop('swimlane_source', None)
        self.swimlane_target = config.pop('swimlane_target', None)

        if self.swimlane_source is None:
            raise  Exception('You must provide a swimlane_source object')

        if self.swimlane_target is None:
            raise  Exception('You must provide a swimlane_target object')

        self.remove_keys = ['createdDate', 'modifiedDate', 'createdByUser', 'modifiedByUser']
        self.mode = config['mode']

        self.dashboard_engine = dashboard(config={'logger':self.logger, 'swimlane_source' : self.swimlane_source, 'swimlane_target' : self.swimlane_target, 'mode' : self.mode})
        self.report_engine = report(config={'logger':self.logger, 'swimlane_source' : self.swimlane_source, 'swimlane_target' : self.swimlane_target, 'mode': self.mode})


    def getWorkspace(self, swimlane_server):
        workspace_response = swimlane_server.swimlane.request(
            'get',
            '/workspaces'
        ).json()
        workspaces = list()
        for workspace in workspace_response:
            workspaces.append(swimlane_server.swimlane.request(
                'get',
                '/workspaces/{}'.format(workspace['id'])
            ).json())

        return workspaces

    def getWorkspaceByApp(self, swimlane_server, appID):
        workspace_response = swimlane_server.swimlane.request(
            'get',
            '/workspaces/app/{}'.format(appID)
        )

        return json.loads(workspace_response.text)


    def addWorkspace(self, swimlane_server, workspace):
        workspace_response = swimlane_server.swimlane.request(
            'post',
            '/workspaces',
            json = workspace
        )

        return json.loads(workspace_response.text)

    def updateWorkspace(self, swimlane_server, workspace_id, group):
        workspace_response = swimlane_server.swimlane.request(
            'put',
            '/workspaces/{}'.format(workspace_id),
            json = group
        )

        return json.loads(workspace_response.text)

    def deleteWorkspace(self, swimlane_server, workspace_id):
        workspace_response = swimlane_server.swimlane.request(
            'delete',
            '/workspaces/{}'.format(workspace_id)
        )


    def removeOrphanWorkspaces(self, swimlane_server):
        workspaces = self.getWorkspace(swimlane_server)
        for workspace in workspaces:

            for applicaiton_id in workspace['applications']:
                application_name = swimlane_server.validateApp(byID=applicaiton_id)
                if application_name == False:
                    workspace['applications'].remove(applicaiton_id)

            if len(workspace['applications']) == 0:
                self.logger.error('Removing Application {}'.format(workspace['name']))
                if self.mode == 'Live':
                    pass
                    # self.deleteWorkspace(swimlane_server, workspace['id'])
        if self.mode == 'Live':
            pass
            # self.dashboard_engine.removeOrphanDashboards(swimlane_server)

    def buildWorkspaceIDs(self, swimlane_server):
        workspaces = self.getWorkspace(swimlane_server)
        workspace_list = []
        for workspace in workspaces:
            workspace_list.append(workspace['id'])

        return workspace_list


    def migrateWorkspaces(self, byAppName = None):
        if byAppName is None:
            source_workspaces = self.getWorkspace(self.swimlane_source)
            target_workspaces = self.getWorkspace(self.swimlane_target)
        else:
            source_app_id = self.swimlane_source.validateApp(byName=byAppName)
            source_workspaces = self.getWorkspaceByApp(self.swimlane_source, source_app_id)
            target_app_id = self.swimlane_target.validateApp(byName=byAppName)
            target_workspaces = self.getWorkspace(self.swimlane_target)


        self.report_engine.migrateReports(self.mode,byApp = {'source' : source_app_id, 'target' : target_app_id})

        for workspace in source_workspaces:
            self.logger.debug('-------------------------------------------------')
            self.logger.info('Working on Workspace {}'.format(workspace['name']))
            target_workspace = {}
            for t_workspace in target_workspaces:
                if t_workspace['name'] == workspace['name']:
                    target_workspace = copy(t_workspace)

            ##Validate all application Workspace is pointing two is on target
            new_application_list = []
            self.logger.debug('## Working On Applications ##')
            for applicaiton_id in workspace['applications']:
                source_application_name = self.swimlane_source.validateApp(byID=applicaiton_id)
                if source_application_name != False:
                    target_application_id = self.swimlane_target.validateApp(byName=source_application_name)
                    if target_application_id != False:
                        if 'applications' not in target_workspace or target_application_id not in target_workspace['applications']:
                            self.logger.info('Add Application {} to Workspace'.format(source_application_name))
                            if self.mode == 'Live':
                                new_application_list.append(target_application_id)

            if len(new_application_list) == 0:
                self.logger.debug('Workspace has valid applications in target SKIPPING Migration')
                #continue

            new_permission_dict = {}
            self.logger.info('## Working On Permissions ##')
            for key, permission in workspace['permissions'].items():
                if key != '$type':
                    if permission['type'] == 'Role':
                        target_role_id = self.swimlane_target.validateRole(byName=permission['name'])
                        if target_role_id != False:
                            self.logger.info('Adding Role {} to Workspace'.format(permission['name']))
                            new_permission_dict[target_role_id] = copy(permission)
                            new_permission_dict[target_role_id]['id'] = target_role_id
                    else:
                        self.logger.critical('Unhandled Permission Type: {}'.format(permission['type']))
                        exit()

            #new_dashboard_list = []

            target_workspace_id_key = find_in_list(target_workspaces, 'name', workspace['name'], self.logger)
            if target_workspace_id_key is None:
                self.logger.critical('Add New')
                for key in self.remove_keys:
                    if key in workspace:
                        workspace.pop(key)

                workspace['applications'] = new_application_list
                workspace['permissions'] = new_permission_dict
                #workspace['dashboards'] = new_dashboard_list

                if self.mode == 'Live':
                    workspace = self.addWorkspace(self.swimlane_target, workspace)
                self.report_engine.migrateReports(self.mode,byWorkspace = {'source' : workspace['id'], 'target' : workspace['id']})
                self.dashboard_engine.migrateDashboards(self.mode,workspace, workspace['id'], self.buildWorkspaceIDs(self.swimlane_target))
            else:
                target_workspace_id = target_workspaces[target_workspace_id_key]['id']

                update_workspace = target_workspaces[target_workspace_id_key]
                update_workspace['applications'] = new_application_list
                for app in target_workspace['applications']:
                    update_workspace['applications'].append(app)
                update_workspace['permissions'] = new_permission_dict

                self.updateWorkspace(self.swimlane_target, update_workspace['id'], update_workspace)
                self.dashboard_engine.migrateDashboards(self.mode,workspace, target_workspace_id, self.buildWorkspaceIDs(self.swimlane_target))











import json, sys

class roles(object):

    def __init__(self, config):

        self.logger = config.pop('logger', None)
        self.swimlane_source = config.pop('swimlane_source', None)
        self.swimlane_target = config.pop('swimlane_target', None)

        if self.swimlane_source is None:
            raise  Exception('You must provide a swimlane_source object')

        if self.swimlane_target is None:
            raise  Exception('You must provide a swimlane_target object')

        self.remove_keys = ['createdDate', 'modifiedDate', 'createdByUser', 'modifiedByUser']
        self.role_exclustions = ['Administrator']
        self.mode = config['mode']


    def getRoles(self, swimlane_server):
        roles_response = swimlane_server.swimlane.request(
            'get',
            '/roles'
        )

        return json.loads(roles_response.text)

    def addRole(self, swimlane_server, role):
        role_response = swimlane_server.swimlane.request(
            'post',
            '/roles',
            json = role
        )

        return json.loads(role_response.text)

    def updateRole(self, swimlane_server, role_id, role):
        role_response = swimlane_server.swimlane.request(
            'put',
            '/roles/{}'.format(role_id),
            json = role
        )

        return json.loads(role_response.text)


    def migrateRoles(self):
        sw_source_roles_raw = self.getRoles(self.swimlane_source)
        sw_target_roles_raw = self.getRoles(self.swimlane_target)

        for role in sw_source_roles_raw:
            if role['name'] not in self.role_exclustions:
                self.logger.info('----------------------------------------')
                self.logger.info('Working on Role: {}'.format(role['name']))
                ## Process Users
                for user_members in role['users']:
                    self.logger.info('Working on User: {}'.format(user_members['name']))
                    source_user_obj = self.swimlane_source.swimlane.users.get(id=user_members['id'])
                    try:
                        target_user_obj = self.swimlane_target.swimlane.users.get(display_name=source_user_obj.display_name)
                        user_members['id'] = target_user_obj.id
                        user_members['name'] = target_user_obj.username
                    except ValueError as e:
                        ## Unable to find user - remove user association
                        self.logger.warn('User: {} not found removing from Role: {}'.format(source_user_obj.display_name, role['name']))
                        if self.mode == 'Live':
                            role['users'].remove(user_members)

                ## Process Groups
                for group_members in role['groups']:
                    self.logger.info('Working on Group: {}'.format(group_members['name']))
                    source_group_obj = self.swimlane_source.swimlane.groups.get(id=group_members['id'])
                    try:
                        target_group_obj = self.swimlane_target.swimlane.groups.get(name=source_group_obj.name)
                        group_members['id'] = target_group_obj.id
                        group_members['name'] = target_group_obj.name
                    except ValueError as e:
                        ## Unable to find group attempt to add
                        self.logger.warn('Group: {} not found attempting to add'.format(source_user_obj.name))
                        if self.mode == 'Live':
                            new_group = {}
                            new_group['id'] = source_group_obj.id
                            new_group['description'] = source_group_obj.description
                            new_group['name'] = source_group_obj.name
                            new_group['disabled'] =  source_group_obj.disabled
                            new_group = addGroup(self.swimlane_target, new_group)

                target_role_key = self.find_in_list(sw_target_roles_raw, 'name', role['name'])
                self.logger.info('Target Key: {}'.format(target_role_key))
                if target_role_key is None:
                    ## New Group
                    for key in self.remove_keys:
                        if key in role:
                            role.pop(key)

                    ## Had to wrap in try it seems that some groups don't get returned when you ask for the groups
                    if self.mode == 'Live':
                        add_role = self.addRole(self.swimlane_target, role)
                        sw_target_roles_raw.append(add_role)

                else:
                    update_role = sw_target_roles_raw[target_role_key]
                    update_role['users'] = role['users']
                    update_role['groups'] = role['groups']
                    if self.mode == 'Live':
                        self.updateRole(self.swimlane_target, update_role['id'], update_role)

    def find_in_list(self, lst, key, value):
        for i, dic in enumerate(lst):
            if key in dic and dic[key] == value:
                return i
        return None


class fields(object):

    def __init__(self, field_obj, config):

        self.field_obj = field_obj
        self.logger = config.pop('logger', None)
        self.swimlane_source = config.pop('swimlane_source', None)
        self.swimlane_target = config.pop('swimlane_target', None)

        if self.logger is None:
            raise Exception('You must provide a logger object')

        if self.swimlane_source is None:
            raise Exception('You must provide a swimlane_source object')

        if self.swimlane_target is None:
            raise Exception('You must provide a swimlane_target object')

        self.failed_fields = []

        self.updates_made = False

        self.logger.debug('Working on Field: {}'.format(field_obj['name']))